/*
 * gfx.c
 * Where all the action takes place.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "config.h"
#include "sprite.h"
#include "proto.h"


extern int fps;
extern int wepnum[];

int **back;	/* The background. */

/* Some colour maps. */
static int blues[257];
static int reds[257];
static int grays[256];
static int fires[257];
static int blast[257];

static int borcol;



/* Draw game border. */
void draw_border()
{
  int x, y;

  /* Draw horizontal borders. */
  for(x = WIDTH / 2; x < RWIDTH - WIDTH / 2; x++) {
    back[x][HEIGHT / 2] = borcol;
    back[x][RHEIGHT - HEIGHT / 2] = borcol;
  }
  
  /* Draw vertical borders. */
  for(y = HEIGHT / 2; y < RHEIGHT - HEIGHT / 2; y++) {
    back[WIDTH / 2][y] = borcol;
    back[RWIDTH - WIDTH / 2][y] = borcol;
  }
}


/* Generate a gradient. */
void gradient(int dir)
{
  int ctr1, ctr2, x, y;
  double cd, cur = 0;
  
  if(dir == 0) {
    x = RWIDTH;
    y = RHEIGHT;
  } else {
    x = RHEIGHT;
    y = RWIDTH;
  }
  
  cd = 256.0 / (double)x;
  
  for(ctr1 = 0; ctr1 < x; ctr1++) {
    for(ctr2 = 0; ctr2 < y; ctr2++)
      back[ctr1][ctr2] = blues[(int)cur];
    
    cur += cd;
  }
}


/* Generate background. */
void mangle()
{
  int x, y, dec, which = random() % BACKS;

#ifdef VERBOSE
  fprintf(stderr, "Generating random background (%d)... ", which);
#endif  

  /* Generate the randomly selected background scheme. */
  switch(which) {

    /* Weird mountainrange thing. */
  case 0:
    for(x = 0; x < RWIDTH; x++)
      back[x][0] = 0;

    for(y = 1; y < RHEIGHT; y++)
      back[0][y] = 0;

    for(x = 1; x < RWIDTH - 1; x++) {
      for(y = 1; y < RHEIGHT - 1; y++) {
	back[x][y] = max2(back[x-1][y], back[x][y-1]);

	if(back[x][y] > 200)
	  dec = 14;
	else
	  dec = 9;
	
	if((back[x][y] += random() % 13 - dec) < 0)
	  back[x][y] = 0;
      }
    }
    break;

    /* Checkerboard. */
  case 1:
    for(x = 0; x < RWIDTH; x++)
      for(y = 0; y < RHEIGHT; y++) {
	if((x / 64) % 2 == (y / 64) % 2)
	  back[x][y] = blues[5];
      else
	back[x][y] = blues[54];
      }
    break;

    /* Starfield. */
  case 2:
    for(x = 0; x < RWIDTH; x++)
      for(y = 0; y < RHEIGHT; y++) {
	if(random() % 128 == 0)
	  back[x][y] = grays[random() % 256];
	else
	  back[x][y] = 0;
      }
    break;

    /* Gradient. */
  case 3:
    gradient(random() % 1);
   
    break;
  }

  /* Draw the game border. */
  draw_border();

#ifdef VERBOSE
  fprintf(stderr, "done.\n");
#endif
}


int ticks = 0;

int xc,        yc;		/* top left corner */
int xd,        yd;		/* course/speed */
int cx,        cy;		/* mouse coords */
int myX,       myY;		/* our absolute coords */
int hisX,      hisY;		/* enemy absolute coords */
int myBombX,   hisBombX;	/* bomb coordinates */
int myBombY,   hisBombY;
int myBombT,   haveBomb,  bombBlast;       /* bomb indicators */

int ehit;
int eHasBomb;
int deathslide, edeathslide;
int state,     estate;
int energy,    eenergy;
int wins = 0,  losses = 0;


/* Initialize a new game. */
void initgame()
{
  deathslide = 0;
  edeathslide = 0;
  state = STATE_ALIVE;
  estate = STATE_ALIVE;
  xc = random() % (RWIDTH - WIDTH);
  yc = random() % (RHEIGHT - HEIGHT);
  xd = 0; yd = 0;
  energy = ENERGY_FULL;
  eenergy = ENERGY_FULL;
  
  ehit = 0;
  myBombT = BOMB_COUNT;
  bombBlast = 0;
  haveBomb = 0;
  eHasBomb = 0;

  wep_removeall();

  myX = xc + WIDTH / 2;
  myY = yc + HEIGHT / 2;

  hisX = RWIDTH / 2;
  hisY = RHEIGHT / 2;

#ifdef VERBOSE
  fprintf(stderr, "Starting new game (wins: %d, losses: %d, rating: %d%%)\n",
	  wins, losses, wins * 100 / max2(wins + losses, 1)); 
#endif
}


/* Redraw bombs. */
void update_bomb(int which)
{
  int x, y, locX, locY, *col, pix;

  if(which == 0) {
    if(!haveBomb) {
      myBombT++;
      return;
    }
    
    if(myBombT > 0)
      myBombT--;

    locX = myBombX;
    locY = myBombY;
    
    if(bombBlast) {
      col = blast;
      bombBlast = 0;
      haveBomb = 0;
    } else 
      col = blues;
 
  } else {
    if(!eHasBomb)
      return;
    locX = hisBombX;
    locY = hisBombY;
    col = reds;
  }
 
  if(locX < xc || locX > xc + WIDTH - BOMBW ||
     locY < yc || locY > yc + HEIGHT - BOMBH)
    return;
  
  for(x = 0; x < BOMBW; x++)
    for(y = 0; y < BOMBH; y++)
      if((pix = bomb[y * BOMBW + x]) != TRANSPARENT)
	gui_put_pixel(locX + x - xc, locY + y - yc, col[pix]);
}


static int energycol[101];
static int energybor;

/* Update our energy graph. */
void update_energy()
{
  int x, y, col = energycol[energy * 100 / ENERGY_FULL],
    en = energy * ENERGY_LEN / ENERGY_FULL;
  
  /* Draw the horizontal borders. */
  for(x = 0; x < ENERGY_LEN + 4; x++) {
    gui_put_pixel(ENERGY_X + x - 2, ENERGY_Y - 2, energybor);
    gui_put_pixel(ENERGY_X + x - 2, ENERGY_Y + 1 + ENERGY_WIDTH, energybor);
  }

  /* Draw the vertical borders. */
  for(y = 0; y < ENERGY_WIDTH + 4; y++) {
    gui_put_pixel(ENERGY_X - 2, ENERGY_Y + y - 2, energybor);
    gui_put_pixel(ENERGY_X + ENERGY_LEN + 1, ENERGY_Y + y - 2, energybor);
  }

  /* Draw the energy bar. */
  for(x = 0; x < en; x++)
    for(y = 0; y < ENERGY_WIDTH; y++)
      gui_put_pixel(ENERGY_X + x, ENERGY_Y + y, col); 
}


/* Update our enemy's energy graph. */
void update_eenergy()
{
  int x, y, col = energycol[eenergy * 100 / ENERGY_FULL],
    en = eenergy * ENERGY_LEN / ENERGY_FULL;

  /* Draw the horizontal borders. */
  for(x = 0; x < ENERGY_LEN + 4; x++) {
    gui_put_pixel(ENERGY_X + x - 2, EENERGY_Y - 2, energybor);
    gui_put_pixel(ENERGY_X + x - 2, EENERGY_Y + 1 + ENERGY_WIDTH, energybor);
  }

  /* Draw the vertical borders. */
  for(y = 0; y < ENERGY_WIDTH + 4; y++) {
    gui_put_pixel(ENERGY_X - 2, EENERGY_Y + y - 2, energybor);
    gui_put_pixel(ENERGY_X + ENERGY_LEN + 1, EENERGY_Y + y - 2, energybor);
  }

  /* Draw the energy bar. */
  for(x = 0; x < en; x++)
    for(y = 0; y < ENERGY_WIDTH; y++)
      gui_put_pixel(ENERGY_X + x, EENERGY_Y + y, col); 
}


/* Update our weps graph. */
void update_weps()
{
  int x, y, len, col;

  len = wepnum[0] * WEPS_LEN / MAX_WEPS;
  col = reds[wepnum[0] * 255 / MAX_WEPS];

  /* Draw the horizontal borders. */
  for(x = 0; x < WEPS_LEN + 4; x++) {
    gui_put_pixel(WEPS_X + x - 2, WEPS_Y - 2, energybor);
    gui_put_pixel(WEPS_X + x - 2, WEPS_Y + 1 + WEPS_WIDTH, energybor);
  }

  /* Draw the vertical borders. */
  for(y = 0; y < WEPS_WIDTH + 4; y++) {
    gui_put_pixel(WEPS_X - 2, WEPS_Y + y - 2, energybor);
    gui_put_pixel(WEPS_X + WEPS_LEN + 1, WEPS_Y + y - 2, energybor);
  }

  /* Draw the energy bar. */
  for(x = 0; x < len; x++)
    for(y = 0; y < WEPS_WIDTH; y++)
      gui_put_pixel(WEPS_X + x, WEPS_Y + y, col); 
}


/* Redraw ourself. */
void update_me()
{
  int x, y, pix;
  
  if(state == STATE_ALIVE) {
    for(x = 0; x < BALLW; x++)
      for(y = 0; y < BALLH; y++)
	if((pix = ball[y * BALLW + x]) != TRANSPARENT)
	  gui_put_pixel(myX + x - xc, myY + y - yc, blues[pix]);    
  } else {
    if(deathslide < DEATHSLIDES) {
      for(x = 0; x < BALLW; x++)
	for(y = 0; y < BALLH; y++)
	  if((pix = death[deathslide][y * BALLW + x]) != TRANSPARENT)
	    gui_put_pixel(myX + x - xc, myY + y - yc, fires[pix]);
    }
    if(++deathslide >= RESTART_DELAY * fps) {
      losses++;
      net_put("R:");
      initgame();
    }
  }
}


/* Redraw the enemy. */
void update_enemy()
{
  int x, y, *col, pix;

  if(ehit) {
    col = blast;
    ehit = 0;
  } else
    col = reds;
 
  if(hisX < xc || hisX > xc + WIDTH - BALLW ||
     hisY < yc || hisY > yc + HEIGHT - BALLH)
    return;

  if(estate == STATE_ALIVE) {
    for(x = 0; x < BALLW; x++)
      for(y = 0; y < BALLH; y++)
	if((pix = ball[y * BALLW + x]) != TRANSPARENT)
	  gui_put_pixel(hisX + x - xc, hisY + y - yc, col[pix]);    
  } else {
    if(edeathslide < DEATHSLIDES) {
      for(x = 0; x < BALLW; x++)
	for(y = 0; y < BALLH; y++)
	  if((pix = death[edeathslide][y * BALLW + x]) != TRANSPARENT)
	    gui_put_pixel(hisX + x - xc, hisY + y - yc, fires[pix]);
      edeathslide++;
    }
  }
}


int lagfactor = 0;		/* we use this to improve sync */

/* Update our netpal. */
void update_pal()
{
  char buf[BUFLEN];

  /* Try to compensate for lag... */
  sprintf(buf, "C:%d:%d:\n", myX + xd * lagfactor, myY + yd * lagfactor);
  net_put(buf);
}


static struct timeval pingtv;
static int lag = 0;

/* Sync us with pal. */
void update_us(char *buf)
{
  char newbuf[BUFLEN];
  char *p = strtok(buf, SEP);
  char *arg1, *arg2, *arg3, *arg4, *arg5, *arg6;
  
  switch(*p) {
  case 'C':
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);

    hisX = atoi(arg1);
    hisY = atoi(arg2);

    /* Enforce boundaries for the enemy's coordinates. */
    if(hisX < 0) hisX = 0;
    else if(hisX >= RWIDTH) hisX = RWIDTH-1;

    if(hisY < 0) hisY = 0;
    else if(hisY >= RHEIGHT) hisY = RHEIGHT-1;

    break;
  case 'M':
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);
    arg3 = strtok(NULL, SEP);
    arg4 = strtok(NULL, SEP);
    arg5 = strtok(NULL, SEP);
    arg6 = strtok(NULL, SEP);

    /* Add this active missile to our collection. */
    wep_new(atoi(arg1), atoi(arg2), atoi(arg3), atoi(arg4),
	    atoi(arg5), atoi(arg6));
    break;
  case 'B':
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);
    
    if(!eHasBomb) {
      eHasBomb = 1;
      wep_newbomb(atoi(arg1), atoi(arg2), 1);
    }
    break;
  case 'H':
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);
    
    ehit = 1;
    eenergy = atoi(arg2);
    wep_rmcode(atoi(arg1));
    break;
  case 'X':
    eHasBomb = 0;
    break;
  case 'P':
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);
    sprintf(newbuf, "E:%s:%s:\n", arg1, arg2);
    net_put(newbuf);
    break;
  case 'E':
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);

    gettimeofday(&pingtv, NULL);

    /* Calculate lag and lagfactor. Important for delay compensation. */
    lag = (pingtv.tv_sec - atol(arg1)) * 1000000 +
      pingtv.tv_usec - atol(arg2);
    lagfactor = (lag * fps) / 1000000;
    if(--lagfactor < 0)
      lagfactor = 0;
    
#ifdef VERBOSE
    if(lag > LAG_ALARM)
      fprintf(stderr, "Network congestion! Lag: %dms\n", lag / 1000);
#endif
     break;
  case 'D':
    eenergy = 0;
    estate = STATE_DEAD;
    break;
  case 'R':
    wins++;
    initgame();
    break;
  }
}


static int mapbor;

/* Redraw the map on our back. */
void update_map()
{
  int x, y;
  
  /* Draw map. */
  for(x = 1; x < MAP_WIDTH - 1; x++)
    for(y = 1; y < MAP_HEIGHT - 1; y++)
      gui_put_pixel(MAP_X + x, MAP_Y + y,
	 back[(x * RWIDTH) / MAP_WIDTH][(y * RHEIGHT) / MAP_HEIGHT]);
  
  /* Draw the borders. */
  for(x = 0; x < MAP_WIDTH; x++) {
    gui_put_pixel(MAP_X + x, MAP_Y, mapbor);
    gui_put_pixel(MAP_X + x, MAP_Y + MAP_HEIGHT, mapbor);
  }
  for(y = 0; y < MAP_HEIGHT; y++) {
    gui_put_pixel(MAP_X, MAP_Y + y, mapbor);
    gui_put_pixel(MAP_X + MAP_WIDTH, MAP_Y + y, mapbor);
  }
  
  /* Draw a dot in our position. */
  if(state == STATE_ALIVE)
    gui_put_pixel(MAP_X + myX * MAP_WIDTH / RWIDTH,
		  MAP_Y + myY * MAP_HEIGHT / RHEIGHT, grays[255]);
  
  /* And one in the enemy's position. */
  if(estate == STATE_ALIVE)
    gui_put_pixel(MAP_X + hisX * MAP_WIDTH / RWIDTH,
		  MAP_Y + hisY * MAP_HEIGHT / RHEIGHT, reds[255]);  
}


extern int wepcols[WEP_COLS];

/* Generate colours. */
void colgen()
{
  int i;

  energybor = gui_rgb(ENERGY_BORDERCOL);
  mapbor = gui_rgb(MAP_BORDERCOL);
  borcol = gui_rgb(BORDER_COLOUR);

  wepcols[0] = gui_rgb(WEP_COL0);
  wepcols[1] = gui_rgb(WEP_COL1);
  wepcols[2] = gui_rgb(WEP_COL2);

  for(i = 0; i < 256; i++)
    blues[i] = gui_rgb(0, MAX_RGB * i / (256 * 16), MAX_RGB * i / 256);
  blues[256] = gui_rgb(50000, 50000, 41000);

  for(i = 0; i < 256; i++)
    reds[i] = gui_rgb(MAX_RGB * i / 256, 0, 0);
  reds[256] = gui_rgb(50000, 50000, 41000);

  for(i = 0; i < 256; i++) {
    grays[i] = gui_rgb(MAX_RGB * i / 255,
		       MAX_RGB * i / 255,
		       MAX_RGB * i / 255);
  }

  for(i = 0; i < 256; i++) {
    fires[i] = gui_rgb(MAX_RGB * (255 - i) / 255,
		       MAX_RGB * (255 - i) / 255,
		       0);
  }
  fires[256] = gui_rgb(50000, 50000, 41000);  

  for(i = 0; i < 257; i++)
    blast[i] = -1;

  /* 100 levels for energy bar */
  for(i = 0; i < 101; i++) {
    if(i > 50) 
      energycol[i] = gui_rgb(MAX_RGB * (100 - i) / 50, MAX_RGB, 1024);
    else
      energycol[i] = gui_rgb(MAX_RGB, MAX_RGB * i / 50, 1024);
  }
}


/* Do any necessary initialization here. */
int gfx_init()
{
  int i;

  /* Allocate memory for the background. */
  if((back = (int **)malloc(RWIDTH * sizeof(int *)))==NULL)
    return 0;
  
  for(i = 0; i < RWIDTH; i++) {
    if((back[i] = (int *)malloc(RHEIGHT * sizeof(int)))==NULL)
      return 0;
  }

  colgen();

  return 1;
}


/* Ping our partner. */
void ping()
{
  char buf[BUFLEN];

  gettimeofday(&pingtv, NULL);
  
  sprintf(buf, "P:%ld:%ld:\n", pingtv.tv_sec, pingtv.tv_usec);
  net_put(buf);
}


/* Make the game go. */
void update()
{
  int mouse;

  /* Keep some kinda track of time. */
  ticks++;

  /* Ping our partner periodically. */
  if(ticks % (PING_FREQ + 1) == PING_BEGIN)
    ping();

  /* Update our screen coordinates. */
  xc += xd; yc += yd;

  /* Enforce horizontal boundaries. */
  if(xc >= RWIDTH - WIDTH)
    xc = RWIDTH - WIDTH - 1;
  else if(xc < 0)
    xc = 0;

  /* Enforce vertical boundaries. */
  if(yc >= RHEIGHT - HEIGHT)
    yc = RHEIGHT - HEIGHT - 1;
  else if(yc < 0)
    yc = 0;

  /* Update the background. */
  gui_update_back(xc, yc, back);

  /* Fire if mouse clicked. */
  if(state == STATE_ALIVE) {
    if((mouse = gui_get_mouse(&cx, &cy)) == 1) {
      wep_new(myX, myY, (cx - WIDTH / 2) / WEP_ACCEL,
	      (cy - HEIGHT / 2) / WEP_ACCEL, WEP_ENERGY, -1);
    } else if(mouse == 2) {
      if(!haveBomb) {
	if(myBombT >= BOMB_COUNT) {
	  wep_newbomb(myX, myY, 0);
	  haveBomb = 1;
	}
      } else if(myBombT <= 0) {
	bombBlast = 1;
	wep_explode();
      }
    }
  }

  /* Update course/speed according to mouse position. */
  xd = (cx - WIDTH / 2) / ME_ACCEL;
  yd = (cy - HEIGHT / 2) / ME_ACCEL;
  
  /* Update our global coordinates. */
  myX = WIDTH/2 + xd + xc;
  myY = HEIGHT/2 + yd + yc;

  update_bomb(0);		/* Update bombs. */
  update_bomb(1);

  update_me();			/* Update our sprite. */
  update_enemy();		/* Draw the enemy, if around. */

  wep_update(0);		/* Update active weapons. */

  update_energy();		/* Update our energy graph. */
  update_eenergy();		/* Update our enemy's energy. */
  update_weps();		/* Update our weapons graph. */
  update_map();                 /* Update the field map. */

  gui_refresh();		/* Redraw the window. */
}



