/*
 * rob.c
 * The Zloe Robot
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "proto.h"


extern int fps, ticks, estate, state;
extern int myX, myY, hisX, hisY;
int difficulty = 0;
int gravity = 0, gravupdate = 1;
int hisPrevX = 0, hisPrevY = 0;


void (*rob_act)(void);


/* Restart the game. */
void rob_restart()
{
  net_put("R:\n");

  initgame();
}


/* Pretend to negotiate a pace. */
void rob_negotiate()
{
  int rate;
  char buf[BUFLEN], *arg1, *arg2;

  net_get(buf);
  strtok(buf, SEP);
  arg1 = strtok(NULL, SEP);
  arg2 = strtok(NULL, SEP);
 
  rate = atoi(arg2);
  sprintf(buf, "I:%s:%d:\n", VERSION, rate);
  net_put(buf);

  fps = 1000000 / rate;
}


/* Action handler for 0 difficulty level. */
void rob_act0()
{
  myX += (hisX - myX) / 18;
  myY += (hisY - myY) / 18;
  
  if(ticks % 10 == 0 && ticks > 0 && estate == STATE_ALIVE)
    
    wep_new(myX, myY, (hisX + fps * (hisX - hisPrevX) - myX) / 16,
	    (hisY + fps * (hisY - hisPrevY) - myY) / 16,
	    WEP_ENERGY, -1);

  hisPrevX = hisX;
  hisPrevY = hisY;
}


/* Action handler for 1 difficulty level. */
void rob_act1()
{
  myX += (hisX - myX) / 14;
  myY += (hisY - myY) / 14;
  
  if(ticks % 6 == 0 && ticks > 0 && estate == STATE_ALIVE)
    
    wep_new(myX, myY, (hisX + fps * (hisX - hisPrevX) - myX) / 16,
	    (hisY + fps * (hisY - hisPrevY) - myY) / 16,
	    WEP_ENERGY, -1);

  hisPrevX = hisX;
  hisPrevY = hisY;
}


/* Action handler for 2 difficulty level. */
void rob_act2()
{
  int distance, lookahead, shootfreq;

  if(ticks % gravupdate == 0) {
    gravity = random() % 100 - 99;
    gravupdate = random() % fps + fps / 2;
  }
  
  distance = min2(abs(hisX - myX), abs(hisY - myY));
  
  /* Run around. */
  myX += (hisX - myX + gravity) / 20; 
  myY += (hisY - myY + gravity) / 20;
  
  if(myX < WIDTH / 2) myX = WIDTH / 2;
  else if(myX > RWIDTH - WIDTH / 2) myX = RWIDTH - WIDTH / 2;

  if(myY < HEIGHT / 2) myY = HEIGHT / 2;
  else if(myY > RHEIGHT - HEIGHT / 2) myY = RHEIGHT - HEIGHT / 2;
    
  shootfreq = distance / (WIDTH / 20) + 1;
  if(shootfreq > ROB_SHOOTFREQ)
    shootfreq = ROB_SHOOTFREQ;
  
  if(distance > 10 &&
     ticks % shootfreq == 0 && ticks > 0 && estate == STATE_ALIVE) {

    if(distance < 32)
      lookahead = fps / 3;
    else
      lookahead = fps * 2 / 3;
    
    wep_new(myX, myY, (hisX + lookahead * (hisX - hisPrevX) - myX) / 16,
	    (hisY + lookahead * (hisY - hisPrevY) - myY) / 16,
	    WEP_ENERGY, -1);
  }

  hisPrevX = hisX;
  hisPrevY = hisY;
}


/* Perform any necessary initialization. */
void rob_init(int dif)
{
  difficulty = dif;

  switch(dif) {
  case 0:
    rob_act = rob_act0;
    break;
  case 1:
    rob_act = rob_act1;
    break;
  case 2:
    rob_act = rob_act2;
    break;
  }
}


/* Make the robot go. */
void rob_update()
{
  ticks++;

  /* Do stuff. */
  rob_act();

  if(ticks == 0 && state == STATE_DEAD) {
    rob_restart();
  }
  
  wep_update(1);
}



