#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "config.h"
#include "sprite.h"
#include "proto.h"


struct wep_list {
  struct wep_list *prev;
  int x, y;
  int dx, dy;
  int energy;
  int code;
  int owner;

  struct wep_list *next;
} *weps = NULL;


int wepnum[2];	      /* the number of active weps for us and our enemy */
int wepcols[WEP_COLS];
extern int image[RWIDTH][RHEIGHT];
extern int fps, xc, yc, myX, myY, energy, state;
extern int myBombX, myBombY, myBombT, hisBombX, hisBombY;
extern int ticks, lagfactor;


/* Drop a bomb. */
void wep_newbomb(int x, int y, int whose)
{
  char buf[BUFLEN];

  if(whose == 0) {
    myBombT = BOMB_COUNT;
    myBombX = x;
    myBombY = y;
    
    sprintf(buf, "B:%d:%d:%d:\n", x, y, BOMB_COUNT);
    net_put(buf);
  } else {
    hisBombX = x;
    hisBombY = y;
  }
}


/* Explode our bomb. */
void wep_explode()
{
  int x, y;
  char buf[BUFLEN];

  /* Notify peer of explosion. */
  sprintf(buf, "X:");
  net_put(buf);

  for(x = -2; x < 3; x++)
    for(y = -2; y < 3; y++)
      wep_new(myBombX, myBombY, 2 * x, 2 * y, WEP_ENERGY, -1);
}


/* Create a new missile with given coordinates and course/speed. */
void wep_new(int x, int y, int dx, int dy, int energy, int code)
{
  struct wep_list *new;
  char buf[BUFLEN];

  /* Check if we have already maxed out on our number of active weps. */
  if(code < 0) {
    if(wepnum[0] >= MAX_WEPS)
      return;
  } else {
    if(wepnum[1] >= MAX_WEPS)
      return;
  }

  new = (struct wep_list *)malloc(sizeof(struct wep_list));
  new->x = x + (BALLW - WEPW) / 2;
  new->y = y + (BALLH - WEPH) / 2;
  new->dx = dx;
  new->dy = dy;
  new->energy = energy;
  new->next = weps;
  new->prev = NULL;

  weps = new;
  if(new->next != NULL)
    new->next->prev = new;

  /* Notify our partner if needed. */
  if(code >= 0) {
    new->code = code;
    new->owner = 1;
  } else {
    new->code = random();

    /* We try to compensate for lag here. */
    sprintf(buf, "M:%d:%d:%d:%d:%d:%d:\n", x + dx * lagfactor,
	    y + dy * lagfactor, dx, dy, new->energy - lagfactor, new->code);
    net_put(buf);
    new->owner = 0;
  }

  wepnum[new->owner]++;
}


/* This missile is burned out. Get rid of it. */
void wep_remove(struct wep_list *p)
{
  if(p->next != NULL)
    p->next->prev = p->prev;
  if(p->prev != NULL)
    p->prev->next = p->next;
  else
    weps = p->next;

  wepnum[p->owner]--;

  free(p);
}


/* Remove all missiles. */
void wep_removeall()
{
  while(weps != NULL)
    wep_remove(weps);
}


/* Remove a missile with the given code. */
void wep_rmcode(int code)
{
  struct wep_list *p = weps;

  while(p != NULL) {
    if(p->code == code) {
      wep_remove(p);
      return;
    }
    p = p->next;
  }
}


/* Draw a missile at the given coordinates. */
void wep_draw(int x, int y, const int *sprite)
{
  int a, b, pix;

  for(a = x; a < x + WEPW; a++)
    for(b = y; b < y + WEPH; b++)
      if((pix = sprite[(b-y) * WEPW + (a-x)]) != -1)
	gui_put_pixel(a - xc, b - yc, wepcols[pix]);
}


/* Update missiles. */
void wep_update(int rob)
{
  struct wep_list *pold, *p = weps;
  char buf[BUFLEN];

  /* Loop for each active missile. */
  while(p != NULL) {

    /* Update coordinates. */
    p->x += p->dx;
    p->y += p->dy;

    /* Make the missile go away if it burns out or hits a boundary. */
    if(--(p->energy) <= 0 || p->x <= WIDTH/2 || p->y <= HEIGHT/2 ||
       p->x >= RWIDTH - WIDTH/2 || p->y >= RHEIGHT - HEIGHT/2) {
      pold = p;
      p = p->next;
      wep_remove(pold);
      continue;
    }

    /* Did this missile hit us? */
    if(p->energy < WEP_ACTIVATE &&
       abs(p->x + WEPW / 2 - myX - BALLW / 2) < WEP_THRESH &&
       abs(p->y + WEPH / 2 - myY - BALLH / 2) < WEP_THRESH) {

      /* We're dead... Schedule a new game. */
      if((energy -= p->energy / WEP_HIT_FRAC + 1) <= 0) {
	state = STATE_DEAD;
	if(rob)
	  ticks = -1 * RESTART_DELAY * fps;

	net_put("D:\n");
      } else {
	sprintf(buf, "H:%d:%d:\n", p->code, energy);
	net_put(buf);
      }
      
      if(!rob)
	wep_draw(p->x, p->y, wep0);

      pold = p;
      p = p->next;
      wep_remove(pold);
      continue;
    }

    /* If this missile is in our visible region, draw it. */
    if(!rob && p->x > xc && p->x < xc + WIDTH - WEPW &&
       p->y > yc && p->y < yc + HEIGHT - WEPH) {
      if(p->energy % 2 == 0)
	wep_draw(p->x, p->y, wep1);
      else
	wep_draw(p->x, p->y, wep2);
    }

    p = p->next;
  }
}







